/* 
  This file is subject to the terms and conditions of the GNU
  General Public License.
  See "http://www.gnu.org/licenses/gpl.txt" for more details.

  Copyright (C) 2004 USUDA Hisashi, All Rights Reserved.
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "ap_alloc.h"

#include <libpq-fe.h>
#include <libpq/libpq-fs.h>

#ifdef DEBUG
#include "http_log.h"
#define DBGLOG(r, fmt, arg...) \
	ap_log_rerror(APLOG_MARK, APLOG_NOTICE, r, fmt, ## arg)
#else
#define DBGLOG(r, fmt, arg...)
#endif
#define SRVLOG(s, fmt, arg...)

struct pgsample_cfgs {
	char *host;
	char *dbname;
	char *user;
	char *table;
	char *id;
	char *type;
	char *object;
};

#define	DEFAULT_HOST_NAME	"localhost"
#define	DEFAULT_TABLE_NAME	"image"
#define	DEFAULT_ID_NAME		"id"
#define DEFAULT_TYPE_NAME	"type"
#define	DEFAULT_OBJECT_NAME	"object"

module MODULE_VAR_EXPORT pgsample_module;

static PGconn *connect_db(request_rec *r, struct pgsample_cfgs *cfg)
{
	PGconn *conn;

	DBGLOG(r, "connect_db: connecting to %s...", cfg->host);
	conn = PQsetdbLogin(cfg->host, NULL, NULL, NULL,
				cfg->dbname, cfg->user, NULL);
	if(PQstatus(conn) == CONNECTION_BAD) {
		return NULL;
	}
	DBGLOG(r, "connect_db: connected (to %s)", cfg->host);

	return conn;
}

static void disconnect_db(PGconn *conn)
{
	PQfinish(conn);
	conn = NULL;
}

static char *check_id(request_rec *r, char *v)
{
	regex_t *reg;
	int ret = 1;

	reg = ap_pregcomp(r->pool, "^[0-9]+$", REG_EXTENDED|REG_NOSUB);
	if(reg) {
		ret = ap_regexec(reg, v, 0, NULL, 0);
	}
	ap_pregfree(r->pool, reg);
	return ret ? NULL:v;
}

static char *get_id(request_rec *r)
{
	char *args = r->args;
	char *p, *k, *v;

	if(!args) {
		return NULL;
	}

	p = ap_getword(r->pool, &args, '&');
	if(p) {
		k = ap_getword(r->pool, &p, '=');
		DBGLOG(r, "get_id: %s -> key=%s", p, k ? k:"<NULL>");
		if(!strcmp(k, "id")) {
			v = ap_getword(r->pool, &p, '=');
			ap_unescape_url(v);
			return check_id(r, v);
		}
	}
	return NULL;
}

static char *get_data(request_rec *r, struct pgsample_cfgs *cfg,
			PGconn *conn, Oid *oidp)
{
	char *cmd;
	char *id, *type, *obj;
	char *ret;
	PGresult *res;

	if((id = get_id(r)) == NULL) {
		return NULL;
	}

	cmd = ap_psprintf(r->pool,
			  "SELECT %s,%s FROM %s WHERE %s = %s",
			  cfg->type, cfg->object, cfg->table,
			  cfg->id, id);
	DBGLOG(r, "get_data: cmd=%s", cmd);
	res = PQexec(conn, cmd);
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		PQclear(res);
		return NULL;
	}

	type  = PQgetvalue(res, 0, 0);
	obj   = PQgetvalue(res, 0, 1);
	DBGLOG(r, "get_data: type=%s object=%s", type, obj);
	if(!type || !obj) {
		return NULL;
	}
	*oidp = strtol(obj, NULL, 10);
	ret   = ap_pstrdup(r->pool, type);
	PQclear(res);
	return ret;
}

static int output_image(request_rec *r, PGconn *conn, Oid oid)
{
	char buf[65536];
	int fd;
	int l;

	PQclear(PQexec(conn, "begin"));
	fd = lo_open(conn, oid, INV_READ);
	if(fd < 0) {
		PQclear(PQexec(conn, "end"));
		return -1;
	}

	while(1) {
		l = lo_read(conn, fd, buf, sizeof(buf));
		if(l <= 0) {
			break;
		}
		ap_rwrite(buf, l, r);
	}
	lo_close(conn, fd);
	PQclear(PQexec(conn, "end"));
	return 0;
}

/***/

static void *pgsample_cdir_cfg(pool *pool, char *arg)
{
	struct pgsample_cfgs *cfg;

	cfg = ap_pcalloc(pool, sizeof(struct pgsample_cfgs));
	cfg->host   = ap_pstrdup(pool, DEFAULT_HOST_NAME);
	cfg->table  = ap_pstrdup(pool, DEFAULT_TABLE_NAME);
	cfg->id     = ap_pstrdup(pool, DEFAULT_ID_NAME);
	cfg->type   = ap_pstrdup(pool, DEFAULT_TYPE_NAME);
	cfg->object = ap_pstrdup(pool, DEFAULT_OBJECT_NAME);
	return (void *)cfg;
}

static int pgsample_handler(request_rec *r)
{
	struct pgsample_cfgs *cfg;
	PGconn *conn;
	Oid oid;
	char *type;

	cfg = ap_get_module_config(r->per_dir_config, &pgsample_module);
	DBGLOG(r, "get_data: cfg=%s %s %s %s %s %s %s",
		cfg->host ? cfg->host:"<null>",
		cfg->dbname ? cfg->dbname:"<null>",
		cfg->user ? cfg->user:"<null>",
		cfg->table ? cfg->table:"<null>",
		cfg->id ? cfg->id:"<null>",
		cfg->type ? cfg->type:"<null>",
		cfg->object ? cfg->object:"<null>");

	if((conn = connect_db(r, cfg)) == NULL) {
		return FORBIDDEN;
	}

	if((type = get_data(r, cfg, conn, &oid)) == NULL) {
		disconnect_db(conn);
		return NOT_FOUND;
	}

	r->content_type = type;
	ap_send_http_header(r);
	if(!r->header_only) {
		output_image(r, conn, oid);
	}

	disconnect_db(conn);
	return OK;
}

/***/

static const command_rec pgsample_cmds[] = {
    {
	"PGsampleHost",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, host),
	OR_ALL, TAKE1,
	"Set Hostname for PostgreSQL server (default:" DEFAULT_HOST_NAME ")",
    },
    {
	"PGsampleDBname",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, dbname),
	OR_ALL, TAKE1,
	"Set DB name for PostgreSQL server",
    },
    {
	"PGsampleUser",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, user),
	OR_ALL, TAKE1,
	"Set Username for PostgreSQL server",
    },
    {
	"PGsampleTable",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, table),
	OR_ALL, TAKE1,
	"Set Table name (default:" DEFAULT_TABLE_NAME ")",
    },
    {
	"PGsampleIdField",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, id),
	OR_ALL, TAKE1,
	"Set Field name of id (default:" DEFAULT_ID_NAME ")",
    },
    {
	"PGsampleTypeField",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, type),
	OR_ALL, TAKE1,
	"Set Field name of mime type (default:" DEFAULT_TYPE_NAME ")",
    },
    {
	"PGsampleObjectField",
	ap_set_string_slot, (void *)XtOffsetOf(struct pgsample_cfgs, object),
	OR_ALL, TAKE1,
	"Set Field name of image object (default:" DEFAULT_OBJECT_NAME ")",
    },
    { NULL },
};

/* Dispatch list of content handlers */
static const handler_rec pgsample_handlers[] = { 
    { "pgsample", pgsample_handler }, 
    { NULL, NULL }
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT pgsample_module = {
    STANDARD_MODULE_STUFF, 
    NULL,                  /* module initializer                  */
    pgsample_cdir_cfg,     /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    pgsample_cmds,         /* table of config file commands       */
    pgsample_handlers,     /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};

